use async_channel::Sender;
use async_std::task;
use glib::Object;
use gtk::{prelude::*, subclass::prelude::*};
use std::{
    cell::RefCell,
    sync::{Arc, RwLock},
    time::Duration,
};

/*
mod sand_area;
use sand_area::{SandArea, CELL_SIZE, HEIGHT, WIDTH};
*/

mod profile;
use profile::profile;

mod physics;
use physics::World;

pub const WIDTH: usize = 601;
pub const HEIGHT: usize = 601;
pub const CELL_SIZE: usize = 1;
const FPS: u64 = 60;

#[derive(Debug)]
pub struct SandViewPrivate {
    rendering: RefCell<bool>,
    last_update: RefCell<std::time::Instant>,
    // area: RefCell<SandArea>,
}

impl Default for SandViewPrivate {
    fn default() -> Self {
        Self {
            rendering: RefCell::new(false),
            last_update: RefCell::new(std::time::Instant::now()),
            // area: RefCell::new(SandArea::default()),
        }
    }
}

#[glib::object_subclass]
impl ObjectSubclass for SandViewPrivate {
    const NAME: &'static str = "SandView";
    type Type = SandView;
    type ParentType = gtk::DrawingArea;
}

impl ObjectImpl for SandViewPrivate {}
impl WidgetImpl for SandViewPrivate {}
impl DrawingAreaImpl for SandViewPrivate {}

glib::wrapper! {
    pub struct SandView(ObjectSubclass<SandViewPrivate>) @extends gtk::DrawingArea, gtk::Widget;
}

impl Default for SandView {
    fn default() -> Self {
        let s: Self = Object::builder().build();
        s.set_width_request(WIDTH as i32);
        s.set_height_request(HEIGHT as i32);

        s.set_draw_func({
            let s = s.clone();
            move |_, context, width, height| {
                profile("redraw", || {
                    context.set_source_rgb(0., 0., 0.);
                    let _ = context.paint();
                    context.set_source_rgb(0.1, 0.1, 0.1);
                    /*
                    for x in (0..width).step_by(CELL_SIZE) {
                        context.move_to(x as f64, 0.);
                        context.line_to(x as f64, HEIGHT as f64);
                    }
                    for y in (0..height).step_by(CELL_SIZE) {
                        context.move_to(0., y as f64);
                        context.line_to(WIDTH as f64, y as f64);
                    }
                    let _ = context.stroke();
                    */
                    /*
                    let area = s.imp().area.borrow();
                    for x in 0..area.width {
                        for y in 0..area.height {
                            if area.grain(x, y) {
                                context.set_source_rgb(0.8, 0.8, 0.8);
                            } else {
                                context.set_source_rgb(0., 0., 0.);
                            }
                            context.rectangle(
                                (x * CELL_SIZE) as f64,
                                (y * CELL_SIZE) as f64,
                                (CELL_SIZE) as f64,
                                (CELL_SIZE) as f64,
                            );
                        }
                    }
                    let _ = context.fill();
                    */
                })
            }
        });

        s
    }
}

impl SandView {
    /*
    fn set_area(&self, area: SandArea) {
        *self.imp().area.borrow_mut() = area;
        self.queue_draw();
    }
    */

    fn set_last_update(&self, now: std::time::Instant) {
        *self.imp().last_update.borrow_mut() = now;
    }

    fn last_update(&self) -> std::time::Instant {
        self.imp().last_update.borrow().clone()
    }
}

/*
async fn animate(sender: Sender<SandArea>) {
    let mut sand_area = SandArea::default();
    sand_area.add_grain(20, 20);
    loop {
        std::thread::sleep(Duration::from_millis(1000 / FPS));
        sand_area = profile("next", || sand_area.next());
        sender.send(sand_area.clone()).await;
    }
}
*/

async fn run_physics(world: World) {
    let mut last_update = std::time::Instant::now();
    let physics_tickspeed = std::time::Duration::from_millis(1000 / 60);
    loop {
        let now = std::time::Instant::now();

        profile("world next", || world.next(now - last_update));
        last_update = wait_for_update(last_update, physics_tickspeed).await;
    }
}

async fn wait_for_update(
    last_update: std::time::Instant,
    tickspeed: std::time::Duration,
) -> std::time::Instant {
    let now = std::time::Instant::now();
    let next_update = last_update + tickspeed;
    task::sleep(next_update - now).await;
    std::time::Instant::now()
}

fn main() {
    let app = gtk::Application::builder()
        .application_id("com.luminescent-dreams.falling-sand")
        .build();

    // let (sender, receiver) = async_channel::bounded(5);

    let world = World::default();

    let runtime = tokio::runtime::Builder::new_multi_thread()
        .enable_all()
        .build()
        .unwrap();

    let physics_thread = runtime.spawn(run_physics(world.clone()));
    /*
        let background_thread = runtime.spawn(animate(sender));
    */

    /*
    let sand_area = Arc::new(RwLock::new(SandArea::default()));
    */

    app.connect_activate(move |app| {
        let window = gtk::ApplicationWindow::new(app);
        window.present();

        let view = SandView::default();
        window.set_child(Some(&view));

        glib::spawn_future_local({
            let view = view.clone();
            view.set_last_update(std::time::Instant::now());
            let fps_delay = std::time::Duration::from_millis(1000 / FPS);
            // let sand_area = sand_area.clone();
            println!("FPS delay: {:?}", fps_delay);

            async move {
                loop {
                    /*
                    profile("setting area", || {
                        view.set_area(sand_area.read().unwrap().clone())
                    });
                    */

                    // Now, determine how much time remains before the next update. This can vary frame by frame, based on how much rendering has to be done.
                    let now = std::time::Instant::now();
                    let mut wait_time = (view.last_update() + fps_delay) - now;
                    if wait_time < std::time::Duration::from_millis(0) {
                        wait_time = fps_delay;
                    }
                    view.set_last_update(now);
                    async_std::task::sleep(wait_time).await;
                }
            }
        });
    });

    app.run();

    // let _ = runtime.block_on(background_thread);
    let _ = runtime.block_on(physics_thread);
}
