use crate::profile;
use std::{
    ops::{Mul, Neg, Sub},
    sync::{Arc, RwLock},
};

const GRAVITY: f64 = 1.;

fn within(val1: f64, val2: f64, tolerance: f64) -> bool {
    (val1 - val2).abs() <= tolerance
}

#[derive(Clone, Copy, Debug, PartialEq)]
struct Point {
    x: f64,
    y: f64,
}

impl Point {
    fn distance_sq(&self, other: &Point) -> f64 {
        let xdist = other.x - self.x;
        let ydist = other.y - self.y;

        xdist * xdist + ydist * ydist
    }

    fn eq(&self, other: &Point, tolerance: f64) -> bool {
        within(self.x, other.x, tolerance) && within(self.y, other.y, tolerance)
    }
}

#[derive(Clone, Copy, Debug, PartialEq)]
struct Vector {
    dx: f64,
    dy: f64,
}

impl Sub for Vector {
    type Output = Vector;
    fn sub(self, rhs: Vector) -> Self::Output {
        Self {
            dx: rhs.dx - self.dx,
            dy: rhs.dy - self.dy,
        }
    }
}

impl Neg for Vector {
    type Output = Vector;
    fn neg(self) -> Self {
        Self {
            dx: -self.dx,
            dy: -self.dy,
        }
    }
}

impl Mul<f64> for Vector {
    type Output = Vector;
    fn mul(self, rhs: f64) -> Self {
        Self {
            dx: self.dx * rhs,
            dy: self.dy * rhs,
        }
    }
}

#[derive(Clone, Copy, Debug, PartialEq)]
struct Grain {
    radius: f64,
    location: Point,
    velocity: Vector,
}

impl Grain {
    fn translate(&mut self, vec: Vector) {
        self.location.x = self.location.x + vec.dx;
        self.location.y = self.location.y + vec.dy;
    }

    fn move_by(&mut self, delta_t: std::time::Duration) {
        self.location.x = self.location.x + self.velocity.dx * (delta_t.as_secs_f64());
        self.location.y = self.location.y + self.velocity.dy * (delta_t.as_secs_f64());
    }
}

#[derive(Clone, Default)]
pub struct World {
    grains: Arc<RwLock<Vec<Grain>>>,
}

impl World {
    pub fn new() -> Self {
        Self {
            grains: Arc::new(RwLock::new(vec![])),
        }
    }

    pub fn add_grain(&self, grain: Grain) {
        self.grains.write().unwrap().push(grain);
    }

    pub fn grain_count(&self) -> usize {
        self.grains.read().unwrap().len()
    }

    pub fn grains(&self) -> Vec<Grain> {
        self.grains.read().unwrap().clone()
    }

    pub fn next(&self, delta_t: std::time::Duration) {
        self.move_objects(delta_t);
        self.resolve_overlaps();
    }

    fn move_objects(&self, delta_t: std::time::Duration) {
        /* First, need to resolve collisions between objects and ensure that they bounce in the correct direction. This will only resolve the initial collision in a step. Additional collisions are undetectable, so we will resolve those with the overlap code. */
        for grain in self.grains.write().unwrap().iter_mut() {
            grain.move_by(delta_t);
        }
    }

    /* Once all movement has been calculated out, some objects are probably still going to be overlapping. */
    fn resolve_overlaps(&self) {}
}

pub trait Collision {
    type Other;
    fn overlap(&self, other: &Self::Other) -> Option<f64>;
    fn resolve_collision(&mut self, other: &mut Self::Other);
}

impl Collision for Grain {
    type Other = Grain;

    fn overlap(&self, other: &Self::Other) -> Option<f64> {
        let radii = (self.radius + other.radius) * (self.radius + other.radius);
        let distance = self.location.distance_sq(&other.location);
        if distance < radii {
            let dx = other.location.x - self.location.x;
            let dy = other.location.y - self.location.y;
            Some(radii.sqrt() - distance.sqrt())
        } else {
            None
        }
    }

    fn resolve_collision(&mut self, other: &mut Self::Other) {
        // between every two objects calculate out the point and angle of incidence for a collision.
        // Start with i, the intersection distance, which is the addition of the radii of both objects.
        // Then calculate t = i / |v1 - v2|
        // I don't think that my equation takes into account two objects that don't collide
        /*
        match self.overlap(other) {
            Some(overlap) => {
                // a collision has occured. First, undo the overlap by moving both pieces backwards to their intersection points
                // once reversed, calculate percentage of travel already done
                // next, calculate the bounce vectors
                // finally, apply the bounce vectors, scaled by remaining travel percentage
            }
            None => {}
        }
        */
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use cool_asserts::assert_matches;

    #[test]
    fn it_detects_grain_grain_collision() {
        let grain_1 = Grain {
            radius: 10.,
            location: Point { x: 0., y: 0. },
            velocity: Vector { dx: 0., dy: 0. },
        };
        let grain_2 = Grain {
            radius: 10.,
            location: Point { x: 1., y: 0. },
            velocity: Vector { dx: 0., dy: 0. },
        };
        let grain_3 = Grain {
            radius: 4.,
            location: Point { x: 15., y: 0. },
            velocity: Vector { dx: 0., dy: 0. },
        };

        assert_matches!(grain_1.overlap(&grain_2), Some(overlap) => {
            assert!(within(overlap, 19., 0.1));
        } );
        assert_matches!(grain_1.overlap(&grain_3), None);
    }

    #[test]
    fn it_moves_a_grain_according_to_time() {
        let mut grain_1 = Grain {
            radius: 10.,
            location: Point { x: 0., y: 0. },
            velocity: Vector { dx: 5., dy: 5. },
        };

        grain_1.move_by(std::time::Duration::from_millis(500));
        let dest = Point { x: 2.5, y: 2.5 };
        assert!(grain_1.location.eq(&dest, 0.001));
    }

    #[test]
    fn it_moves_many_grains() {
        let grain_1 = Grain {
            radius: 5.,
            location: Point { x: 0., y: 0. },
            velocity: Vector { dx: 0., dy: -1. },
        };
        let grain_1_b = Grain {
            location: Point { x: 0., y: -0.5 },
            ..grain_1
        };
        let grain_2 = Grain {
            radius: 5.,
            location: Point { x: 5., y: 5. },
            velocity: Vector { dx: 0., dy: 1. },
        };
        let grain_2_b = Grain {
            location: Point { x: 5., y: 5.5 },
            ..grain_2
        };
        let grain_3 = Grain {
            radius: 5.,
            location: Point { x: 15., y: 0. },
            velocity: Vector { dx: 1., dy: 1. },
        };
        let grain_3_b = Grain {
            location: Point { x: 15.5, y: 0.5 },
            ..grain_3
        };

        let world = World::new();
        world.add_grain(grain_1);
        world.add_grain(grain_2);
        world.add_grain(grain_3);

        world.next(std::time::Duration::from_millis(500));
        let grains = world.grains();
        for grain in grains {
            let tolerance = 0.001;
            assert!(
                grain.location.eq(&grain_1_b.location, tolerance)
                    || grain.location.eq(&grain_2_b.location, tolerance)
                    || grain.location.eq(&grain_3_b.location, tolerance)
            );
        }
    }

    #[test]
    fn it_resolves_a_collision_when_no_collision_is_present() {
        let mut grain_1 = Grain {
            radius: 10.,
            location: Point { x: 0., y: 0. },
            velocity: Vector { dx: 0., dy: 0. },
        };
        let mut grain_2 = Grain {
            radius: 4.,
            location: Point { x: 15., y: 0. },
            velocity: Vector { dx: 0., dy: 0. },
        };

        let grain_1_b = grain_1.clone();
        let grain_2_b = grain_2.clone();

        grain_1.resolve_collision(&mut grain_2);
        assert!(grain_1.location.eq(&grain_1_b.location, 0.001));
        assert!(grain_2.location.eq(&grain_2_b.location, 0.001));
    }

    #[test]
    fn it_resolves_a_collision() {
        let mut grain_1 = Grain {
            radius: 4.,
            location: Point { x: 0., y: 0. },
            velocity: Vector { dx: 0., dy: 0. },
        };
        let mut grain_2 = Grain {
            radius: 4.,
            location: Point { x: 4., y: 0. },
            velocity: Vector { dx: 0., dy: 0. },
        };

        grain_1.resolve_collision(&mut grain_2);
        assert!(
            grain_1.location.eq(&Point { x: -2., y: 0. }, 0.001),
            "{:?}",
            grain_1
        );
        assert!(grain_2.location.eq(&Point { x: 6., y: 0. }, 0.001));
        assert_eq!(grain_1.overlap(&grain_2), None);
    }
}
