/*
Copyright 2023-2024, Savanni D'Gerinel <savanni@luminescent-dreams.com>

This file is part of FitnessTrax.

FitnessTrax is free software: you can redistribute it and/or modify it under the terms of the GNU
General Public License as published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

FitnessTrax is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License along with FitnessTrax. If not, see <https://www.gnu.org/licenses/>.
*/

use crate::types::{
    DistanceFormatter, DurationFormatter, FormatOption, ParseError, TimeFormatter, WeightFormatter,
};
use gtk::prelude::*;
use std::{cell::RefCell, rc::Rc};

type Parser<T> = dyn Fn(&str) -> Result<T, ParseError>;
type OnUpdate<T> = dyn Fn(Option<T>);

#[derive(Clone)]
pub struct TextEntry<T: Clone + std::fmt::Debug> {
    value: Rc<RefCell<Option<T>>>,

    widget: gtk::Entry,
    parser: Rc<Parser<T>>,
    on_update: Rc<OnUpdate<T>>,
}

impl<T: Clone + std::fmt::Debug> std::fmt::Debug for TextEntry<T> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> Result<(), std::fmt::Error> {
        write!(
            f,
            "{{ value: {:?}, widget: {:?} }}",
            self.value, self.widget
        )
    }
}

// I do not understand why the data should be 'static.
impl<T: Clone + std::fmt::Debug + 'static> TextEntry<T> {
    pub fn new<R, V, U>(
        placeholder: &str,
        value: Option<T>,
        renderer: R,
        parser: V,
        on_update: U,
    ) -> Self
    where
        R: Fn(&T) -> String + 'static,
        V: Fn(&str) -> Result<T, ParseError> + 'static,
        U: Fn(Option<T>) + 'static,
    {
        let widget = gtk::Entry::builder().placeholder_text(placeholder).build();
        if let Some(ref v) = value {
            widget.set_text(&renderer(v))
        }

        let s = Self {
            value: Rc::new(RefCell::new(value)),
            widget,
            parser: Rc::new(parser),
            on_update: Rc::new(on_update),
        };

        s.widget.buffer().connect_text_notify({
            let s = s.clone();
            move |buffer| s.handle_text_change(buffer)
        });

        s
    }

    fn handle_text_change(&self, buffer: &gtk::EntryBuffer) {
        if buffer.text().is_empty() {
            *self.value.borrow_mut() = None;
            self.widget.remove_css_class("error");
            (self.on_update)(None);
            return;
        }
        match (self.parser)(buffer.text().as_str()) {
            Ok(v) => {
                *self.value.borrow_mut() = Some(v.clone());
                self.widget.remove_css_class("error");
                (self.on_update)(Some(v));
            }
            // need to change the border to provide a visual indicator of an error
            Err(_) => {
                self.widget.add_css_class("error");
            }
        }
    }

    pub fn widget(&self) -> gtk::Widget {
        self.widget.clone().upcast::<gtk::Widget>()
    }

    #[cfg(test)]
    fn has_parse_error(&self) -> bool {
        self.widget.has_css_class("error")
    }
}

pub fn time_field<OnUpdate>(
    value: Option<TimeFormatter>,
    on_update: OnUpdate,
) -> TextEntry<TimeFormatter>
where
    OnUpdate: Fn(Option<TimeFormatter>) + 'static,
{
    TextEntry::new(
        "HH:MM",
        value,
        |val| val.format(FormatOption::Abbreviated),
        TimeFormatter::parse,
        on_update,
    )
}

pub fn distance_field<OnUpdate>(
    value: Option<DistanceFormatter>,
    on_update: OnUpdate,
) -> TextEntry<DistanceFormatter>
where
    OnUpdate: Fn(Option<DistanceFormatter>) + 'static,
{
    TextEntry::new(
        "0 km",
        value,
        |val| val.format(FormatOption::Abbreviated),
        DistanceFormatter::parse,
        on_update,
    )
}

pub fn duration_field<OnUpdate>(
    value: Option<DurationFormatter>,
    on_update: OnUpdate,
) -> TextEntry<DurationFormatter>
where
    OnUpdate: Fn(Option<DurationFormatter>) + 'static,
{
    TextEntry::new(
        "0 m",
        value,
        |val| val.format(FormatOption::Abbreviated),
        DurationFormatter::parse,
        on_update,
    )
}
pub fn weight_field<OnUpdate>(
    weight: Option<WeightFormatter>,
    on_update: OnUpdate,
) -> TextEntry<WeightFormatter>
where
    OnUpdate: Fn(Option<WeightFormatter>) + 'static,
{
    TextEntry::new(
        "0 kg",
        weight,
        |val| val.format(FormatOption::Abbreviated),
        WeightFormatter::parse,
        on_update,
    )
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::gtk_init::gtk_init;

    fn setup_u32_entry() -> (Rc<RefCell<Option<u32>>>, TextEntry<u32>) {
        let current_value = Rc::new(RefCell::new(None));

        let entry = TextEntry::new(
            "step count",
            None,
            |steps| format!("{}", steps),
            |v| v.parse::<u32>().map_err(|_| ParseError),
            {
                let current_value = current_value.clone();
                move |v| *current_value.borrow_mut() = v
            },
        );

        (current_value, entry)
    }

    #[test]
    fn it_responds_to_field_changes() {
        gtk_init();
        let (current_value, entry) = setup_u32_entry();
        let buffer = entry.widget.buffer();

        buffer.set_text("1");
        assert_eq!(*current_value.borrow(), Some(1));

        buffer.set_text("15");
        assert_eq!(*current_value.borrow(), Some(15));
    }

    #[test]
    fn it_preserves_last_value_in_parse_error() {
        crate::gtk_init::gtk_init();
        let (current_value, entry) = setup_u32_entry();
        let buffer = entry.widget.buffer();

        buffer.set_text("1");
        assert_eq!(*current_value.borrow(), Some(1));

        buffer.set_text("a5");
        assert_eq!(*current_value.borrow(), Some(1));
        assert!(entry.has_parse_error());
    }

    #[test]
    fn it_update_on_empty_strings() {
        gtk_init();
        let (current_value, entry) = setup_u32_entry();
        let buffer = entry.widget.buffer();

        buffer.set_text("1");
        assert_eq!(*current_value.borrow(), Some(1));
        buffer.set_text("");
        assert_eq!(*current_value.borrow(), None);

        buffer.set_text("1");
        assert_eq!(*current_value.borrow(), Some(1));
        buffer.set_text("1a");
        assert_eq!(*current_value.borrow(), Some(1));
        assert!(entry.has_parse_error());

        buffer.set_text("");
        assert_eq!(*current_value.borrow(), None);
        assert!(!entry.has_parse_error());
    }
}
