use chrono::{DateTime, FixedOffset, NaiveDate};
use dimensioned::si;
use emseries::{Recordable, Timestamp};
use serde::{Deserialize, Serialize};

/// SetRep represents workouts like pushups or situps, which involve doing a "set" of a number of
/// actions, resting, and then doing another set.
#[allow(dead_code)]
pub struct SetRep {
    /// I assume that a set/rep workout is only done once in a day.
    date: NaiveDate,
    /// Each set entry represents the number of times that the action was performed in a set. So, a
    /// pushup workout that involved five sets would have five entries. Each entry would be x
    /// number of pushups. A viable workout would be something like [6, 6, 4, 4, 5].
    sets: Vec<u32>,
    comments: Option<String>,
}

/// The number of steps one takes in a single day.
#[derive(Clone, Debug, PartialEq, Serialize, Deserialize)]
pub struct Steps {
    pub date: NaiveDate,
    pub count: u32,
}

impl Recordable for Steps {
    fn timestamp(&self) -> Timestamp {
        Timestamp::Date(self.date)
    }

    fn tags(&self) -> Vec<String> {
        vec![]
    }
}

/// TimeDistance represents workouts characterized by a duration and a distance travelled. These
/// sorts of workouts can occur many times a day, depending on how one records things. I might
/// record a single 30-km workout if I go on a long-distanec ride. Or I might record multiple 5km
/// workouts if I am out running errands. Distance and Duration are both optional because different
/// people have different priorities and may choose to measure different things.
#[derive(Clone, Debug, PartialEq, Serialize, Deserialize)]
pub struct TimeDistance {
    /// The precise time (and the relevant timezone) of the workout. One of the edge cases that I
    /// account for is that a ride which occurred at 11pm on one day in one timezone would then
    /// count as 1am on the next day if the user moves two timezones to the east. While technically
    /// correct, for most users this would throw off many years of metrics in ways that can be very
    /// hard to understand. Keeping the fixed offset means that we can have the most precise time
    /// in the database, but we can still get a Naive Date from the DateTime, which will still read
    /// as the original day.
    pub datetime: DateTime<FixedOffset>,
    /// The distance travelled. This is optional because such a workout makes sense even without
    /// the distance.
    pub distance: Option<si::Meter<f64>>,
    /// The duration of the workout, which is also optional. Some people may keep track of the
    /// amount of distance travelled without tracking the duration.
    pub duration: Option<si::Second<f64>>,
    pub comments: Option<String>,
}

/// A singular daily weight measurement. Weight changes slowly enough that it seems unlikely to
/// need to track more than a single weight in a day.
#[derive(Clone, Debug, PartialEq, Serialize, Deserialize)]
pub struct Weight {
    pub date: NaiveDate,
    pub weight: si::Kilogram<f64>,
}

impl Recordable for Weight {
    fn timestamp(&self) -> Timestamp {
        Timestamp::Date(self.date)
    }

    fn tags(&self) -> Vec<String> {
        vec![]
    }
}

#[derive(Clone, Debug, PartialEq)]
pub enum RecordType {
    BikeRide,
    Row,
    Run,
    Steps,
    Swim,
    Walk,
    Weight,
}

/// The unified data structure for all records that are part of the app.
#[derive(Clone, Debug, PartialEq, Serialize, Deserialize)]
pub enum TraxRecord {
    BikeRide(TimeDistance),
    Row(TimeDistance),
    Run(TimeDistance),
    Steps(Steps),
    Swim(TimeDistance),
    Walk(TimeDistance),
    Weight(Weight),
}

impl TraxRecord {
    pub fn workout_type(&self) -> RecordType {
        match self {
            TraxRecord::BikeRide(_) => RecordType::BikeRide,
            TraxRecord::Row(_) => RecordType::Row,
            TraxRecord::Run(_) => RecordType::Run,
            TraxRecord::Steps(_) => RecordType::Steps,
            TraxRecord::Swim(_) => RecordType::Swim,
            TraxRecord::Walk(_) => RecordType::Walk,
            TraxRecord::Weight(_) => RecordType::Weight,
        }
    }

    pub fn is_weight(&self) -> bool {
        matches!(self, TraxRecord::Weight(_))
    }

    pub fn is_steps(&self) -> bool {
        matches!(self, TraxRecord::Steps(_))
    }
}

impl Recordable for TraxRecord {
    fn timestamp(&self) -> Timestamp {
        match self {
            TraxRecord::BikeRide(rec) => Timestamp::DateTime(rec.datetime),
            TraxRecord::Row(rec) => Timestamp::DateTime(rec.datetime),
            TraxRecord::Run(rec) => Timestamp::DateTime(rec.datetime),
            TraxRecord::Steps(rec) => rec.timestamp(),
            TraxRecord::Swim(rec) => Timestamp::DateTime(rec.datetime),
            TraxRecord::Walk(rec) => Timestamp::DateTime(rec.datetime),
            TraxRecord::Weight(rec) => rec.timestamp(),
        }
    }

    fn tags(&self) -> Vec<String> {
        vec![]
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use emseries::Series;

    #[test]
    fn can_record_records() {
        let file = tempfile::NamedTempFile::new().expect("a temporary file");
        let path = file.into_temp_path();
        let mut series: Series<TraxRecord> = Series::open(&path).unwrap();

        let record = TraxRecord::Steps(Steps {
            date: NaiveDate::from_ymd_opt(2023, 1, 1).unwrap(),
            count: 1000,
        });

        let id = series.put(record.clone()).unwrap();

        let record_ = series.get(&id).unwrap();
        assert_eq!(record_.data, record);
    }
}
